/* $Id: TuningType.java,v 1.8 2011/11/11 08:19:51 mit Exp $ */
/* TuningType.java v0.2 */
import java.util.*;

public class TuningType implements Runnable
{
    private boolean Runsw;
    private int AW; // all wire
    private int OB; // single wound end
    private int DB; // w wound end
    private int SB; // single key start
    private int FB; // start key#
    private int CB; // check ver.
    private int P4th;
    private int P5th;
    private int Oct2;
    private int A37;
    private int A61;
    private int Tsw; // tuning sw.
    private int Bw;  // оkey 
    private int Inum;  // interval
    private int Itype; // inha type
    private int SepKey[];
    private int Sep[];
    private double Beat = 1.0;
    private double Lay = 1.0;
    private double ModE; // 
    private Vector Vgen; // Wire
    private Thread thr;
    private Interval I = new Interval();

    final static int OCT = 12;
    final static int OCT2 = OCT*2;
    final static int MJ3 = 4;  // MJ3rd
    final static int P4 = 5;   // P4th
    final static int P5 = 7;   // P5th
    final static int MI6 = 8;  // mi6th
    final static int NOT = -1; // tune not
    final static int LAY = 0;  // " laying
    final static int EXT = 1;  // " extension
    final static int SPR = 2;  // " separate
    final static int PYT = 3;  // " pythagoras
    final static int OC2 = 4;  // " 2octave
    final static int IH0 = 0;  // " equal
    final static int IHL = 1;  // inha. old type
    final static int IHN = 2;  // " B type
    final static int CENT = 1200;

    public TuningType() { }
    public TuningType(Vector gen, int b, int c, int d, int e, int f) {
	setWireData(gen,b,c,d,e,f);
    }
    public TuningType(Vector gen, int[] h) {
	setWireData(gen, h[0], h[1], h[2], h[3], h[4]);
    }
    public void setWireData(Vector gen, int b, int c, int d, int e, int f) {
	Vgen = gen;
	AW = Vgen.size();
	OB = b;
	DB = c;
	SB = d;
	FB = e;
	CB = f;
	ModE = (CB > 0) ? 1.08*100000 : 2.15*100000;
	P4th = I.getIinte(P4);
        P5th = I.getIinte(P5);
	Oct2 = I.getIinte(OCT2);
	A37 = getK2W(37);
	A61 = getK2W(61);
    }
    public void setItype(int isw) {
	Itype = isw;
    }
    public void setInterval(int itv) {
	Inum = itv;
    }
    public void setSepKey(int[] sep, int[] sepk) {
	Sep = sep;
	SepKey = sepk;
    }
    //
    public void startRun(double be, double ly, int bw, int type) {
	Beat = be;
	Lay = ly;
	Bw = bw;
	Tsw = type;

        thr = new Thread(this);
        thr.start();
	Runsw = true;
    }
    public boolean isRun() {
	return Runsw;
    }
    public void stop() {
        if (thr != null) {
            thr = null;
	    Runsw = false;
	}
    }
    public void run() {
        int stime = 90;
	int fmax = nextKey(Bw, MI6);    // +f
	int fmin = nextKey(Bw, -MJ3);   // -f

	switch (Tsw) {
	case LAY:
	    laying(stime, fmin, fmax);
	    break;
	case EXT: // extension
	case SPR: // separate
	    extension(stime, fmax);
	    break;
	case PYT:
	    pythagoras(fmax);
	    break;
	case OC2:
	    octave2(stime);
	    break;
	default:
	}

	Tsw = NOT;
	Runsw = false;
    }
    //---
    public void laying(int stime, int fmin, int fmax) {
	double bc = getCent(Bw);        // keycent
	double cc = Lay/OCT;            // 1keyʬcentͥХ
	int st = -MJ3;                  // MJ3 -4

	if (getW2K(fmin) == getW2K(fmin-1)) // 2ξ
	    fmin--;                     // disp_lay()Ʊ
	if (fmin < FB) //
	    fmin = FB;

	unison_check();
	for (int i = fmin; i < fmax; i++) { // fmax2ǥɽϥ
	    try {
		if (i == Bw)  // keyѹ
		    continue;
		else
		    setCent(i, bc+cc*st); //gen_set(i, bc+cc*st);

		if (i < fmax-1 && getW2K(i) != getW2K(i+1))
		    st++;     // Ʊ+ʤ

		thr.sleep(stime);
	    } catch (InterruptedException e) {
		System.out.println(e);
	    } 
	}
    }
    public void unison_check() {
	int pw = Bw+1, mw = Bw-1;

	if (getW2K(Bw) == getW2K(pw))
	    setCent(pw, getCent(Bw));
	else if (getW2K(Bw) == getW2K(mw))
	    setCent(mw, getCent(Bw));
    }
    //---
    public void extension(int stime, int fmax) {
	int in, emin = nextKey(Bw, -MJ3-1); // -f-1

	unison_check();
	try {
	    for (int i = fmax; i < AW; i++) { 
		in = (Tsw == SPR) ? get_intr(i) : Inum;
		if (!tune_set(i, in, Beat, true))
		    break;
		thr.sleep(stime);
	    }
	    for (int i = emin; i >= FB; i--) { // 0?
		if (isLength(i))			
		    continue;

		in = (Tsw == SPR) ? get_intr(i) : Inum;
		if (!tune_set(i, in, Beat, false))
		    break;
		thr.sleep(stime);
	    }
	} catch (InterruptedException e) {
	    System.out.println(e);
	}
    }
    public int get_intr(int w) {
	int x, k = getW2K(w)-1;

	if (k < SepKey[1])      // 0~11
	    x = Sep[0];
	else if (k < SepKey[2]) // 12~23
	    x = Sep[1];
	else if (k < SepKey[3]) // 24~47
	    x = Sep[2];
	else if (k < SepKey[4]) // 48~59
	    x = Sep[3];
	else
	    x = Sep[4];         // 60~87

	return x;
    }
    //---
    public void pythagoras(int fmax) {
	int stime = 100;
	int x = OCT, bw = Bw, owr;

	while (x > 0) {
	    try {
		owr = nextKey(bw, I.getIvalue(P4th));
		if (owr >= fmax) {
		    owr = nextKey(bw, -I.getIvalue(P5th));
		    lay_set(bw, owr, false);
		} else
		    lay_set(bw, owr, true);
		bw = owr;
		thr.sleep(stime);
		x--;
	    } catch (InterruptedException e) {
		x = 0;
		System.out.println(e);
	    } 
	}
    }
    public void lay_set(int nw, int nx, boolean sw) {
	int zero = 0;
	int in = sw ? P4th : P5th;
	int_set(nw, nx, in, zero, sw);

	int ii = sw ? nx+1 : nx-1;
	if (getW2K(nx) == getW2K(ii))
	    int_set(nw, ii, in, zero, sw);
    }
    //---
    public void octave2(int stime) {
	try {
	    double sa = get_cent_val();
	    set_2octave(sa/(A61-A37), sa/2);

	    for (int i = A61+1; i < AW; i++) {
		if (!tune_set(i, Oct2, Beat, true))
		    break;
		thr.sleep(stime);
	    }
	    for (int i = A37-1; i >= FB; i--) { // 0?
		if (isLength(i))
		    continue;

		if (!tune_set(i, Oct2, Beat, false))
		    break;
		thr.sleep(stime);
	    }
	} catch (InterruptedException e) {
	    System.out.println(e);
	}
    }
    public double get_cent_val() {
	int k38 = getK2W(38);
	int k62 = getK2W(62);
	Inter k2 = new Inter(k38, k62, Oct2);

	int k36 = getK2W(36);
	int k60 = getK2W(60);
	Inter k1 = new Inter(k36, k60, Oct2); 

	double dc = 0.01, up = 0;
	double ds, sa, sb, d;
	do {
	    ds = up/(A61-A37);
	    set_2octave(ds, up/2);
	    tuneBeats(true, Beat, k2);
	    sa = getCent(k62)-getCent(A61);

	    tuneBeats(false, Beat, k1);
	    sb = getCent(A37)-getCent(k36);

	    d = (sa-ds > 1.0) ? dc*10 : dc;
	    up += d;
	} while (sa > ds && sb > ds);

	return up-d;
    }
    public void set_2octave(double ds, double sf) {
	int n = 0;

	for (int i = A37; i <= A61; i++)
	    setCent(i, ds*(n++)-sf);
    }
    //---
    public boolean tune_set(int nw, int in, double be, boolean sw) {
	int nxt = nextKey(nw, (sw ? -I.getIvalue(in) : I.getIvalue(in)));
	if (nxt < FB || nxt >= AW)
	    return false;

	int_set(nxt, nw, in, be, sw);
	return true;
    }
    public void int_set(int nxt, int now, int in, double be, boolean sw) {
	Inter In = sw ? new Inter(nxt, now, in) : new Inter(now, nxt, in);
	tuneBeats(sw, be, In);
    }
    public void gen_set(int w, double dt) {
	setCent(w, dt);
    }
    public int nextKey(int k, int itr) {
	int x = k+itr;
	if (x < FB || x >= AW)
            return 0;

	int nkey = getW2K(k)+itr;
	int ukey = getW2K(x);
	while (nkey > ukey && ukey < AW)
	    ukey = getW2K(++x);
	while (nkey < ukey && ukey > FB)
	    ukey = getW2K(--x);

	return x;
    }
    public double tuneBeats(boolean sw, double max, Inter k) {
	double d, dc = 0.01;
	double mc = getBeats(k);
	int wn;

	if (sw) {
            wn = k.getHkey();
            dc = dc;
        } else {
            wn = k.getLkey();
            dc = -dc;
        }

	double sa = 0.4;
	while (mc < max) {
	    d = (max-mc > sa) ? dc*10 : dc;
	    addCent(wn, d);
	    mc = getBeats(k);
	}
	while (mc > max) {
	    d = (max-mc > sa) ? dc*10 : dc;
	    addCent(wn, -d);
	    mc = getBeats(k);
	}

	return mc;
    }
    //---
    public double getBeats(Inter n) {
	return getIfreq(n.getHkey(), n.getLint())-
	    getIfreq(n.getLkey(), n.getHint());
    }
    public double getInha(int wn) {
	return (Itype == IHN) ? getBinhar(wn, 1) : getInhar(wn);
    }
    public double getInha(int wn, int pr) {
	switch (Itype) {
	case IHL:
	    return getInhar(wn)*pr*pr;
	case IHN:
	    return getBinhar(wn, pr);
	case IH0:
	default:
	    return 0;
	}
    }
    public double getCent(int wn) {
	return getVcent(wn);
    }
    public double getIcent(int wn, int pr) {
	return getInha(wn, pr)+getCent(wn);
    }
    public double getFreq(int wn) {
	return getVfreq(wn);
    }
    public double getIfreq(int wn, int pr) {
	return ctof(getFreq(wn)*pr, getInha(wn, pr));
    }
    public void addCent(int wn, double ce) {
	addVcent(wn, ce);
    }
    public void setCent(int wn, double c) {
	setVcent(wn, c);
    }
    public int getMinWire() {
	return getVMinWire();
    }
    public int getMaxWire() {
	return getVMaxWire();
    }
    public int getMaxKey() {
	return getVMaxKey();
    }
    public int getK2W(int k) {
	return getVK2W(k);
    }
    public int getW2K(int wn) {
	return getVW2K(wn);
    }
    public double getInhar(int wn) {
	return getVinhar(wn);
    }
    //
    public double getLength(int wn) {
	return getVlength(wn);
    }
    public double getBinhar(int wn, int pa) {
	return getVBinhar(wn, pa);
    }
    public double getB(int wn) {
	return getVB(wn);
    }
    public double getDensity(int wn) { // debug
        return getVdensity(wn);
    }
    public double getGtension(int wn) {
	return getVGtension(wn);
    }
    public double getStress(int wn) {
	return getVstress(wn);
    }
    public double getGauge(int wn) {
	return getVgauge(wn);
    }
    public String toString(int wn) {
	return getVs(wn);
    }
    public String getHeaderS() {
	return getVheaderS();
    }
    public int[] getHeader() {
	return getVheader();
    }
    public boolean isCopper(int wn) {
	return isVcopper(wn);
    }
    public boolean isLength(int wn) {
	return (getVlength(wn) <= 0);
    }
    public boolean isPiano() {
        return isVcsw();
    }
    //
    public Wire getVg(int wn) {
	return (Wire)Vgen.get(wn); // (Wire)Vgen.elementAt(wn);
    }
    public double getVlength(int wn) {
        return getVg(wn).getLength();
    }
    public double getVinhar(int wn) {
        return getVg(wn).getInhar();
    }
    public double getVB(int wn) {
	return getVg(wn).getB(ModE);
    }
    public double getVBinhar(int wn, int p) {
        return getVg(wn).getBInhar(p, ModE);
    }
    public double getVcent(int wn) {
        return getVg(wn).getCent();
    }
    public double getVfreq(int wn) {
        return getVg(wn).getFreq();
    }
    public double getVdensity(int wn) { // debug
        return getVg(wn).getDensity();
    }
    public double getVGtension(int wn) {
        return getVg(wn).getGTension();
    }
    public double getVstress(int wn) {
        return getVg(wn).getStress();
    }
    public double getVgauge(int wn) {
	return getVg(wn).getGauge();
    }
    public String getVs(int wn) {
	return getVg(wn).toString();
    }
    public String getVheaderS() {
	return getVg(0).getHeaderString();
    }
    public int[] getVheader() {
	return getVg(0).getHeader();
    }
    public void addVcent(int wn, double ce) {
        getVg(wn).addCent(ce);
    }
    public void setVcent(int wn, double ce) {
        getVg(wn).setCent(ce);
    }
    public boolean isVcopper(int wn) {
        return getVg(wn).isCopper();
    }
    public int getVMinWire() {
	return getVg(0).getMinWire();
    }
    public int getVMaxWire() {
	return getVg(0).getMaxWire();
    }
    public int getVMaxKey() {
	return getVg(0).getMaxKey();
    }
    public int getVK2W(int k) { // 1~88 -> 0~
	return getVg(0).getK2W(k);
    }
    public int getVW2K(int wn) { // 0~ -> 1~
	return getVg(wn).getW2K();
    }
    public boolean isVcsw() {
        return getVg(0).isCsw();
    }
    //
    public double ctof(double f, double cent) {
	return f*Math.pow(2, cent/CENT);
    }
}
